/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <ox/std/typeinfo.hpp>

#include "descwrite.hpp"

namespace ox {

namespace detail {

struct preloadable_type {
	static constexpr auto Preloadable = true;
};

struct non_preloadable_type {
	static constexpr auto Preloadable = false;
};

struct non_preloadable_type2 {
};

static_assert(preloadable<preloadable_type>::value);
static_assert(!preloadable<non_preloadable_type>::value);
static_assert(!preloadable<non_preloadable_type2>::value);

}


static_assert([] {
	int i = 0;
	return detail::indirectionLevels(i) == 0;
}(), "indirectionLevels broken: indirectionLevels(int)");

static_assert([] {
	int i = 0;
	return detail::indirectionLevels(&i) == 1;
}(), "indirectionLevels broken: indirectionLevels(int*)");

static_assert([] {
	int i[2] = {};
	return detail::indirectionLevels(i) == 1;
}(), "indirectionLevels broken: indirectionLevels(int[])");

const DescriptorType *TypeDescWriter::type(int8_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:int8_t";
	constexpr auto PT = PrimitiveType::SignedInteger;
	constexpr auto Bytes = 1;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(int16_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:int16_t";
	constexpr auto PT = PrimitiveType::SignedInteger;
	constexpr auto Bytes = 2;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(int32_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:int32_t";
	constexpr auto PT = PrimitiveType::SignedInteger;
	constexpr auto Bytes = 4;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(int64_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:int64_t";
	constexpr auto PT = PrimitiveType::SignedInteger;
	constexpr auto Bytes = 8;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(uint8_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:uint8_t";
	constexpr auto PT = PrimitiveType::UnsignedInteger;
	constexpr auto Bytes = 1;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(uint16_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:uint16_t";
	constexpr auto PT = PrimitiveType::UnsignedInteger;
	constexpr auto Bytes = 2;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(uint32_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:uint32_t";
	constexpr auto PT = PrimitiveType::UnsignedInteger;
	constexpr auto Bytes = 4;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(uint64_t*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:uint64_t";
	constexpr auto PT = PrimitiveType::UnsignedInteger;
	constexpr auto Bytes = 8;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(char*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:string";
	constexpr auto PT = PrimitiveType::String;
	return getType(String, PT, 0, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(SerStr, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:string";
	constexpr auto PT = PrimitiveType::String;
	return getType(String, PT, 0, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(String*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:string";
	constexpr auto PT = PrimitiveType::String;
	return getType(String, PT, 0, alreadyExisted);
}

const DescriptorType *TypeDescWriter::type(bool*, bool *alreadyExisted) noexcept {
	constexpr auto String = "B:bool";
	constexpr auto PT = PrimitiveType::Bool;
	constexpr auto Bytes = 0;
	return getType(String, PT, Bytes, alreadyExisted);
}

const DescriptorType *TypeDescWriter::getType(const String &tn, PrimitiveType pt, int b, bool *alreadyExisted) noexcept {
	auto t = m_typeStore->get(tn);
	if (!t.error) {
		*alreadyExisted = true;
		auto type = t.value;
		oxAssert(type != nullptr, "TypeDescWriter::getType returning null DescriptorType");
		return type;
	} else {
		*alreadyExisted = false;
		auto dt = ox::make_unique<DescriptorType>(tn);
		dt->primitiveType = pt;
		dt->length = b;
		const auto out = dt.get();
		m_typeStore->set(tn, std::move(dt));
		return out;
	}
}

}
